#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Dec  4 13:42:45 2019

@author: oomsba
"""
import os
import numpy as np
from netCDF4 import Dataset
import rasterio

#creating cogs
from rasterio.io import MemoryFile
from rio_cogeo.cogeo import cog_translate
from rio_cogeo.profiles import cog_profiles

def get_footprint_value(infile):
    footprint_value = 'MULTIPOLYGON (((-180 %latmax%, -180 %latmin%, 180 %latmin%, 180 %latmax%s, -180 %latmax%)))'
    with rasterio.open(infile) as dataset:
        # Read the dataset's valid data mask as a ndarray.
        mask = dataset.dataset_mask()
        ymax_array = mask.max(axis=1)
        firsthit = 0
        lasthit = 0
        for yindex in range(0, len(ymax_array)):
            if ymax_array[yindex] == 255:
                firsthit = yindex
                break

        for yindex in range(len(ymax_array)-1 ,0, -1):
            if ymax_array[yindex] == 255:
                lasthit = yindex
                break

        last = rasterio.transform.xy(dataset.transform, lasthit, 0)
        latmin = last[1]
        first = rasterio.transform.xy(dataset.transform, firsthit, 0)
        latmax = first[1]
        footprint_value = footprint_value.replace('%latmax%',str(latmax))
        footprint_value = footprint_value.replace('%latmin%',str(latmin))
        return footprint_value


def convert_nc2tiff(input_file, output_dir, tif_settings):
    #some defaults
    latpixels2skip = 20
    tif_files=[]

    #collecting some general settings
    version_major = tif_settings[0]['tif_file_name'][-7]
    version_minor = tif_settings[0]['tif_file_name'][-6]
    version_patch = tif_settings[0]['tif_file_name'][-5]
    tiftagdate = tif_settings[0]['tif_file_name'].split('_')[3]

    for tif_setting in tif_settings:
        scale = tif_setting['scale']
        unit_conversion_factor = tif_setting['unit_conversion_factor']
        datatype = tif_setting['dtype']
        offset = tif_setting['offset']
        #update processingmode in description
        processingmode = tif_setting['processingmode']
        description = tif_setting['metadata']['description'].replace('%processing_mode%', processingmode)
        tif_setting['metadata']['description'] = description

        metadata = tif_setting['metadata']
        nodata =  tif_setting['nodata']
        band_name = tif_setting['nc_band_name']
        tif_name = tif_setting['tif_file_name'].replace('<band>', tif_setting['tif_band_name'])

        d_nc = Dataset(input_file, 'r', format="NETCDF4")

        lat_bounds = d_nc.variables['latitude_bounds'][:]
        lon_bounds = d_nc.variables['longitude_bounds'][:]
        lat = (lat_bounds[0, :, 0] + lat_bounds[0, :, 1]) / 2.
        lon = (lon_bounds[0, :, 0] + lon_bounds[0, :, 1]) / 2.

        # test on variable CO
        d_var = d_nc.variables[band_name][0, :, :]  #blabal 

        #scaling
        # PV = DV * input_scale + PV_offset
        # DV = (PV - PV_offset)/ input_scale
        d_var = d_var / unit_conversion_factor          #e.g from mol to µmol
        d_var_scaled = (d_var - offset)/scale
        d_var_scaled = d_var_scaled.astype(datatype)

        #setting masked data with nodata value
        d_var_scaled_masked = d_var_scaled.filled(nodata)

        #TODO is this needed?
        d_var_scaled_masked = d_var_scaled_masked[::-1]     # reverse array so the tif looks like the array

        # generate the geotransform
        nx = len(lon)
        ny = len(lat)
        xmin, ymin, xmax, ymax = [lon_bounds.min(), lat_bounds.min(), lon_bounds.max(), lat_bounds.max()]
        xres = (xmax - xmin) / float(nx)
        yres = (ymax - ymin) / float(ny)

        geotransform = rasterio.transform.from_origin(-180, 90 - yres*latpixels2skip, xres, yres)   #map the pixels to coordinates

        output_file = os.path.join(output_dir, tif_name)

        profile = {'driver': 'GTiff',
                   'height': d_var_scaled_masked.shape[0]-2*latpixels2skip,
                   'width': d_var_scaled_masked.shape[1],
                   'count': 1,
                   "crs": "epsg:4326",
                   'dtype': str(d_var_scaled_masked.dtype),
                   'transform': geotransform,
                   }

        with MemoryFile() as memfile:
            with memfile.open(**profile) as mem:
                # Populate the input file with numpy array
                # update tif tags
                mem.update_tags(TIFFTAG_COPYRIGHT= 'VITO')
                mem.update_tags(TIFFTAG_DATETIME="%s-%s-%s" % (tiftagdate[0:4], tiftagdate[4:6], tiftagdate[6:8]))
                mem.update_tags(TIFFTAG_SOFTWARE="Workflow:%s.%s.%s" % (version_major, version_minor, version_patch))
                mem.update_tags(units=metadata['units'])
                mem.write(d_var_scaled_masked[latpixels2skip:-latpixels2skip, :], 1)    #remove the data around 90 and -90 degrees
                mem.set_band_description(1, metadata['description'])
                #mem.units='myinit'   #Settings units per band this way doesn't work
                mem.offsets=[offset]
                mem.scales=[scale]
                mem.nodata=nodata
                dst_profile = cog_profiles.get("deflate")  #'deflate': {'driver': 'GTiff', 'interleave': 'pixel', 'tiled': True, 'blockxsize': 512, 'blockysize': 512, 'compress': 'DEFLATE'}
                cog_translate(
                    mem,
                    output_file,
                    dst_profile,
                    in_memory=True,
                    quiet=True,
                )
        print("Done writing %s" % output_file)
        tif_files.append(output_file)
    return 0, tif_files
