import os
from datetime import timedelta, datetime
import re
import copy
import glob

import logging
pyspark_log = logging.getLogger('pyspark').setLevel(logging.ERROR)  # remove anoying pyspark messages
# logger = logging.getLogger("py4j")
# logger.setLevel(logging.INFO)
# logger.addHandler(logging.StreamHandler())


##VERSIONING
RESOLUTION_DICT = {
    "CO": {
        "TD" : 0.05,
        "TM" : 0.05,
        "TY" : 0.05,
    },
    "NO2": {
        "TD" : 0.02,
        "TM" : 0.02,
        "TY" : 0.02,
    },
    "HCHO": {
        "TD" : 0.2,
        "TM" : 0.05,
        "TY" : 0.05,
    },
    "CH4": {
        "TD" : 0.05,
        "TM" : 0.05,
        "TY" : 0.05,
    },
}
TIF_DICT = {
    "CO": [{'nc_band_name': 'CO_column_volume_mixing_ratio_dry_air',  # used fore reading the nc-file
            'tif_band_name': 'CO',  # used in the tif file name
            'unit_conversion_factor' : 1,   #ppbv => no conversion for this band
            'scale': 10,
            "offset": 0,
            "dtype": 'int16',
            "nodata": 32767,
            "metadata": {
                "description": "Averaged carbon monoxide column volume mixing ratio for dry air [ppbv] based on Sentinel-5P/TROPOMI Level-2 %processing_mode% data.",
                "units": "[ppbv]"
            }
            },
           {'nc_band_name': 'weight',
            'tif_band_name': 'WEIGHT',  # used in the tif file name
            'unit_conversion_factor': 1,   #convert from mol to mol => no conversion for this band
            'scale': 1,
            "offset": 0,
            "dtype": 'int16',
            "nodata": -9999,
            "metadata": {
                "description": "weighting factors used for spatial binning based on Sentinel-5P/TROPOMI Level-2 %processing_mode% data.",
                "units": "-"
            }}
           ],
    "NO2": [{'nc_band_name': 'tropospheric_NO2_column_number_density',
             'tif_band_name': 'NO2',  # used in the tif file name
             'unit_conversion_factor' : 1e-6,   #convert from mol to µmol
             'scale': 5,
             "offset": 0,
             "dtype": 'int16',
             "nodata": 32767,
             "metadata": {
                 "description": "Tropospheric vertical column of nitrogen dioxide [µmol m^-2] based on Sentinel-5P/TROPOMI Level-2 %processing_mode% data.",
                 "units": "[µmol m^-2]"
             }},
            {'nc_band_name': 'weight',
             'tif_band_name': 'WEIGHT',  # used in the tif file name
             'unit_conversion_factor' : 1,   #convert from mol to mol => no conversion for this band
             'scale': 1,
             "offset": 0,
             "dtype": 'int16',
             "nodata" : -9999,
             "metadata": {
                 "description": "weighting factors used for spatial binning based on Sentinel-5P/TROPOMI Level-2 %processing_mode% data.",
                 "units": "-"
             }}
            ],
    "HCHO": [{'nc_band_name': 'tropospheric_HCHO_column_number_density',
              'tif_band_name': 'HCHO',  # used in the tif file name
              'unit_conversion_factor': 1e6,    #unit in nc_file is mol/m^2 => requested is µmol/m² # d_var = d_var / unit_conversion_factor
              #convert from mol/m2 to molecules/m2: Use Avogadro constant 6.02214e23
              #convert from molecule/m2 => molecules /cm²              : 6.02214e19
              #convert from molecule/cm² => molecules /cm²(*10^15)     : 6.02214e4
              #in the formule we do a division with the scaling factor : 1/6.02214e4
              'scale': 0.02,  # min value 0, max_value ?600?= > max digital value is 30000
              "offset": 0,
              "dtype": 'int16',
              "nodata": -9999,  # TODO
              "metadata": {
                  "description": "Tropospheric formaldehyde column number density [µmol m^-2] based on Sentinel-5P/TROPOMI Level-2 %processing_mode% data.",
                  "units": "[µmol m^-2]"
              }},
             {'nc_band_name': 'weight',
              'tif_band_name': 'WEIGHT',  # used in the tif file name
              'unit_conversion_factor': 1,  # convert from mol to mol => no conversion for this band
              'scale': 1,
              "offset": 0,
              "dtype": 'int16',
              "nodata": -9999,
              "metadata": {
                  "description": "weighting factors used for spatial binning based on Sentinel-5P/TROPOMI Level-2 %processing_mode% data.",
                  "units": "-"
              }}
             ],
    "CH4": [{'nc_band_name': 'CH4_column_volume_mixing_ratio_dry_air',
             'tif_band_name': 'CH4',  # used in the tif file name
             'unit_conversion_factor' : 1,   #ppbv => no conversion for this band
             'scale': 0.05,             #min 1650 -> 2000
             "offset": 1650,            # PV = DV * input_scale + PV_offset      # DV = (PV - PV_offset)/ input_scale
             "dtype": 'int16',
             "nodata": -9999,
             "metadata": {
                 "description": "Tropospheric vertical column of methane [ppb] based on Sentinel-5P/TROPOMI Level-2 %processing_mode% data.",
                 "units": "[ppb]"
             }},
            {'nc_band_name': 'weight',
             'tif_band_name': 'WEIGHT',  # used in the tif file name
             'unit_conversion_factor' : 1,   #convert from mol to mol => no conversion for this band
             'scale': 1,
             "offset": 0,
             "dtype": 'int16',
             "nodata" : -9999,
             "metadata": {
                 "description": "weighting factors used for spatial binning based on Sentinel-5P/TROPOMI Level-2 %processing_mode% data.",
                 "units": "-"
             }}
            ]
}
class s5p_L3_product():
    def __init__(self, input_root_dir, output_dir_nc, output_dir_tif, process_date, product_type, composit_type, processing_mode, version, use_rpro=False):
        #### Definitions
        #if logger is None:
        self.logger = logging.getLogger()
        self.status=1  #when processed successfull, it wil be 0
        self.nr_of_inputs = 0
        self.input_root_dir = input_root_dir
        self.version = version              # e.g 100, 200
        self.out_detailed_version = "000000"
        self.use_rpro = use_rpro    #when True we will search for RPRO L2-products too
        self.latstep = RESOLUTION_DICT[product_type][composit_type]
        self.lonstep = self.latstep
        # if composit_type == 'TD' and product_type == 'HCHO':
        #     self.latstep = 0.2
        #     self.lonstep = 0.2
        self.keep = 'keep(%s,latitude_bounds,longitude_bounds,datetime_start)'  # variables to keep when loading S5P file: only these are loaded into memory, %s is filled later on with the right target variable
        self.exclude_default = 'exclude(latitude_bounds_weight,longitude_bounds_weight,latitude,longitude)'
        self.oper_box = []
        self.oper_bin = 'bin_spatial(%d,-90,%.2f,%d,-180,%.2f)' % (
            int(180 / self.latstep) + 1, self.latstep, int(360 / self.lonstep + 1), self.lonstep)
        self.outfolder_nc = None
        self.outfolder_tif = None
        self.begin_position = None     #begin position of the first segment, this is a datetime object
        self.end_position = None        #end position of the last segment, this is a datetime object
        self.keep_in_files_lst = []    #this list will contain the valid input files only

        self.process_date = process_date    # e.g 20200101  #first day of the product yyyymmdd
        process_date_obj = datetime.strptime(process_date, '%Y%m%d')
        self.process_end_date = None        # last day of the product e.g 20201231 for a TY product yyyymmdd
        self.composit_type = composit_type
        self.processing_level = 'L3'
        if composit_type == 'TD':                                       #Daily product
            self.process_end_date = self.process_date                    #e.g 20220101
        elif composit_type == 'TM':                                     #Monthly Product
            self.process_end_date = datetime.strftime((process_date_obj.replace(day=1) + timedelta(days=32)).replace(day=1) - timedelta(days=1), '%Y%m%d') # transform to start of the month, add 32 days, transform to the first day of next month again, substract 1 day and we will get the last day of the period
        elif composit_type == 'TY':                                     #yearly products
            self.process_end_date = datetime.strftime((process_date_obj.replace(day=1) + timedelta(days=366)).replace(day=1) - timedelta(days=1), '%Y%m%d')

        self.revision_number = '0%s' % version[0]           # 100 -> 01, 200-> 02
        self.product_type = product_type                       #e.g CO/NO2
        self.processing_mode = processing_mode                #e.g NRTI/OFFL
        self.collection = 'S5P_%s_%s_%s_%s_%s' %(processing_mode, self.processing_level, product_type, composit_type, version)     #S5P_NRTI_L3_CO_TD_200
        self.title = 'S5P_%s_L3_%s_%s_%s_V%s' % (       #S5P_OFFL_L3_NO2_TD_20200316_V100
            self.processing_mode,
            self.product_type,
            self.composit_type,
            self.process_date,
            self.version
        )

        ##### INPUTS #######
        # list available products
        # if minimal_version is None:  # if minimal version is empty, we use the revision number to remove version with a major difference in version numbering
        #     minimal_version = '%s0000' % self.revision_number  # eg. 020000

        if composit_type == 'TD':
            self.status = self._collect_L2_inputs(composit_type=self.composit_type, minimal_version_nr='0%s0000' % version[0], use_rpro=self.use_rpro)   #V200 will result in minimal version 020000
        elif composit_type == 'TM':
            if self.product_type == 'HCHO':
                self.status = self._collect_L2_inputs(composit_type=self.composit_type, minimal_version_nr='0%s0000' % version[0], use_rpro=self.use_rpro)  # only for HCHO TM will be based on L2-data
            else:
                self.status = self._collect_L3_inputs(composit_type=self.composit_type, minimal_version_nr='0%s0000' % version[0])  # V200 will result in minimal version 020000
        elif composit_type == 'TY':
            self.status = self._collect_L3_inputs(composit_type=self.composit_type, minimal_version_nr='0%s0000' % version[0])  # V200 will result in minimal version 020000

        if self.status != 0:
            print ("No valid inputs are found")
            return
        ##### OUTPUTS#####
        self._set_outfilename( output_dir_nc, output_dir_tif )
        self._create_output_folder(output_dir_nc, output_dir_tif )

        ### CREATE HARPMERGE command #####
        self._set_process_parameters()

    def set_status(self, status):
        #status 0: OK, -1 : FAILED, 1:NODATA
        self.status = status

    def _last_day_of_month(self, yyyymmdd):
        start_date = datetime.strptime(yyyymmdd,'%Y%m%d')
        # The day 28 exists in every month. 4 days later, it's always next month
        next_month = start_date.replace(day=28) + timedelta(days=4)
        # subtracting the number of the current day brings us back one month
        last_day =  next_month - timedelta(days=next_month.day)
        return last_day.strftime('%Y%m%d')

    def _add_day(self, yyyymmdd):
        date_obj = datetime.strptime(yyyymmdd,'%Y%m%d')
        date_obj = date_obj + timedelta(days=1)
        return date_obj.strftime('%Y%m%d')

    def _collect_L2_inputs(self, minimal_version_nr='000000', composit_type='TD', use_rpro=False):
        #tests/data/input/2020/03/16/S5P_OFFL_L2__NO2____20200316T225229_20200317T003400_12564_01_010302_20200319T144311/S5P_OFFL_L2__NO2____20200316T225229_20200317T003400_12564_01_010302_20200319T144311.nc"
        yyyymmdd = self.process_date
        self.input_processing_level = 'L2'

        if use_rpro:    #use RPRO data too
            processing_mode_pattern = '[RO][PF][RF][OL]' #search for RPRO + OFFL
        else:
            processing_mode_pattern = self.processing_mode

        if composit_type == 'TD':
            yyyy = yyyymmdd[0:4]; mm = yyyymmdd[4:6]; dd = yyyymmdd[6:8]

            pattern = os.path.join(self.input_root_dir, yyyy, mm, dd, 'S5P_%s_%s__%s_%s*/S5P_%s_%s__%s_%s*.nc' % (
                processing_mode_pattern, self.input_processing_level, self.product_type.ljust(6,'_'),
                yyyymmdd, processing_mode_pattern, self.input_processing_level,
                self.product_type.ljust(6,'_'), yyyymmdd))

            #for TD (just one date)
            all_files = glob.glob(pattern)  # collect ALL files from FS, without taking version into account

        elif composit_type == 'TM':
            #for TM collect files for all dates
            last_of_month = self._last_day_of_month(yyyymmdd)
            date_2_collect = yyyymmdd
            all_files=[]
            while date_2_collect <= last_of_month:  #loop over each day for that month to collect the L2-data
                yyyy = date_2_collect[0:4]
                mm = date_2_collect[4:6]
                dd = date_2_collect[6:8]
                pattern = os.path.join(self.input_root_dir, yyyy, mm, dd, 'S5P_%s_%s__%s_%s*/S5P_%s_%s__%s_%s*.nc' % (
                    processing_mode_pattern, 'L2', self.product_type.ljust(6, '_'),
                    date_2_collect, processing_mode_pattern, 'L2',
                    self.product_type.ljust(6, '_'), date_2_collect))
                date_files = glob.glob(pattern)  # collect ALL files from FS, without taking version into account
                all_files.extend(date_files)
                date_2_collect = self._add_day(date_2_collect)  #increment day

        #all_files now contains all unfiltered paths for TD or TM
        if len(all_files) == 0:
            self.logger.error("The provide pattern doesn't match with any file: %s" % pattern)
            return 4

        #remove files with unwanted versions, or double orbits
        orbit_dict = {}                 #create a dict having only one product per orbit to prevent using double overpasses. An overpass is defined by the sensing_key(begin_end_orbitnr).
                                        #N.B : OFFL/RPRO L2 data have 1 file per overpass, while NRTI have several segments per overpass.
        skip_counter_minimal_version = 0

        #building orbit dict, so double orbits can be filtered
        for fpath in all_files:
            fname = os.path.basename(fpath)
            fname = fname.replace("__", "_").replace("__", "_").replace("__", "_")    #removing all double '_', also ___ -> '_'
            fname = fname.replace(".nc", '')
            try: #removing fileextention
                s5p, in_processing_mode, in_level, in_ptype, in_beginposition, in_endposition, in_orbitnr, in_revision, in_version, in_processing_time = fname.split("_")
            except:
                raise Exception ("Unable to parse filename %s into [s5p, in_processing_mode, in_level, in_ptype, in_beginposition, in_endposition, in_orbitnr, in_revision, in_version, in_processing_time]" % fname)


            #skip unwanted input files
            in_version_major = in_version[0:2]
            minimal_version_nr_major = minimal_version_nr[0:2]
            if in_version_major != minimal_version_nr_major:
                print("Skipping files due to version_nr: %s < %s" % (in_version, minimal_version_nr))
                skip_counter_minimal_version += 1
                continue

            #only use 1 file per orbit number
            sensing_key = '%s_%s_%s' % (in_beginposition, in_endposition, in_orbitnr)
            if not sensing_key in orbit_dict.keys():
                orbit_dict[sensing_key] = fpath
            else:
                #orbit already present, user RPRO version here
                if in_processing_mode == 'RPRO':
                    orbit_dict[sensing_key] = fpath  #overwrite by RPRO

        #orbit_dict only contains one file per orbit, unwanted versions are already filtered
        keep_in_files_lst = []  # the list of files that will be kept for output processing
        stimes = [];
        etimes = []  # get lowest starttime and highest endtime from inputs
        for sensing_key, fpath in orbit_dict.items():
            fname = os.path.basename(fpath)
            fname = fname.replace("__", "_").replace("__", "_").replace("__","_")  # removing all double '_', also ___ -> '_'
            fname = fname.replace(".nc", '')
            s5p, in_processing_mode, in_level, in_ptype, in_beginposition, in_endposition, in_orbitnr, in_revision, in_version, in_processing_time = fname.split("_")
            #keep track of lowest beginposition and highest endposition
            stimes.append(datetime.strptime(in_beginposition, '%Y%m%dT%H%M%S'))
            etimes.append(datetime.strptime(in_endposition, '%Y%m%dT%H%M%S'))
            #keep track of highest input version
            if in_version > self.out_detailed_version:                        #keep highest input version e.g 010400
                self.out_detailed_version = in_version
            keep_in_files_lst.append(fpath)             #these files are valid

        if len(keep_in_files_lst) > 0:
            self.begin_position = min(stimes).strftime('%Y%m%dT%H%M%S')     #get final time from all used inputs
            self.end_position = max(etimes).strftime('%Y%m%dT%H%M%S')    #get final time for all used inputs
            self.nr_of_inputs = len(keep_in_files_lst)
            self.infile_str = ' '.join(keep_in_files_lst)   #convert the list into a string which can be used directly in the harpmerge command
        else:
            print("Files found in the directory : %d" % len(all_files))
            print("Files skipped due to minimal_version : %d" % skip_counter_minimal_version)
            self.logger.error("There are no files for processing")
            return 3
        return 0


    def _collect_L3_inputs(self, minimal_version_nr='000000', composit_type='TM'):

        self.input_processing_level = 'L3'
        yyyymmdd = self.process_date
        yyyy = yyyymmdd[0:4]; mm = yyyymmdd[4:6]; dd = yyyymmdd[6:8]


        #TY and TM
        #by default, TD is used for TY and TM products
        mm_pattern = mm                     #for file lookup only take files for mm into account
        if self.composit_type == 'TY':
            mm_pattern = '*'                #for TY products, we need all TD files for ALL months
        #input_root_dir = /data/TERRASCOPE/s5p/output_nc/L3_NO2_TD_V1/
        pattern = os.path.join(self.input_root_dir, yyyy, mm_pattern, 'S5P_%s_L3_%s_%s_%s%s*_V%s*/S5P_OFFL_L3%s%s_*.nc' % (
                                         self.processing_mode, self.product_type, 'TD',
                                         yyyy, mm_pattern, self.version[0], 'TD', self.product_type))   #always use TD, also for TY products
        #Use TM for HCHO TY(the resolution for TD is to low here)
        if self.composit_type == 'TY' and self.product_type == 'HCHO':
            pattern = os.path.join(self.input_root_dir, yyyy,
                                   'S5P_%s_L3_%s_%s_%s%s01_V%s*/S5P_OFFL_L3%s%s_*.nc' % (
                                       self.processing_mode, self.product_type, 'TM',
                                       yyyy, mm_pattern, self.version[0], 'TM',
                                       self.product_type))  # always use TD, also for TY products
        all_files = glob.glob(pattern)  # collect ALL files from FS, without taking version into account

        if len(all_files) == 0:
            self.logger.error("The provide pattern doesn't match with any file: %s" % pattern)
            return 4

        #remove files with unwanted versions
        keep_in_files_lst = []          #the list of files that will be kept for output processing
        stimes = []; etimes = []        #get lowest starttime and highest endtime from inputs
        skip_counter_minimal_version = 0
        for fpath in all_files:
            fname = os.path.basename(fpath)
            fname = fname.replace("__", "_").replace("__", "_").replace("__", "_")    #removing all double '_', also ___ -> '_'
            fname = fname.replace(".nc", '')                                          #removing fileextention

            #parse filename in parts
            fname_parts = fname.split("_")
            if len(fname_parts) == 7:
                s5p, in_processing_mode, in_level_TD_ptype, in_beginposition, in_endposition, in_counts, in_version = fname_parts  #S5P_OFFL_L3TDNO2___20230104T000359_20230105T000719_000c3_020400.nc
            elif len(fname_parts) == 8:
                s5p, in_processing_mode, in_level_TD_ptype, in_beginposition, in_endposition, in_counts, in_collections, in_version = fname_parts  # S5P_OFFL_L3TDNO2___20230104T000359_20230105T000719_000c3_02_020400.nc
            else:
                raise Exception ("Unable to parse filename %s into [s5p, in_processing_mode, in_level_TD_ptype, in_beginposition, in_endposition, in_counts, in_collection, in_version]" % fname)

            #skip unwanted input files
            if in_processing_mode != self.processing_mode:
                continue
            # if in_revision != self.revision_number:   #skip mixing 01 and 02 revision
            #     continue                              #test disabled e.g S5P_OFFL_L2__CO_____20221101T224837_20221102T003006_26184_03_020400_20221103T123224.nc
            if in_version < minimal_version_nr:
                print("Skipping files due to version_nr: %s < %s" % (in_version, minimal_version_nr))
                skip_counter_minimal_version += 1
                continue

            #keep track of lowest beginposition and highest endposition
            stimes.append(datetime.strptime(in_beginposition, '%Y%m%dT%H%M%S'))
            etimes.append(datetime.strptime(in_endposition, '%Y%m%dT%H%M%S'))
            #keep track of highest input version
            if in_version > self.out_detailed_version:                        #keep highest input version e.g 010400
                self.out_detailed_version = in_version
            keep_in_files_lst.append(fpath)             #these files are valid

        if len(keep_in_files_lst) > 0:
            self.begin_position = min(stimes).strftime('%Y%m%dT%H%M%S')     #get final time from all used inputs
            self.end_position = max(etimes).strftime('%Y%m%dT%H%M%S')    #get final time for all used inputs

            self.begin_position = '%sT000000' % self.process_date        #reset timings to fixed ones for TM|TY
            self.end_position = '%sT235959' % self.process_end_date      #reset timings to fixed ones for TM|TY
            self.nr_of_inputs = len(keep_in_files_lst)
            self.infile_str = ' '.join(keep_in_files_lst)   #convert the list into a string which can be used directly in the harpmerge command
        else:
            print("Files found in the directory : %d" % len(all_files))
            print("Files skipped due to minimal_version : %d" % skip_counter_minimal_version)
            self.logger.error("There are no files for processing")
            return 3
        return 0

    def _set_outfilename(self, output_dir_nc, output_dir_tif):
        #creating directory and file names
        #collection = '%s_%s_%s_V%s' % (self.processing_level, self.product_type, self.composit_type, self.version)  #L3_NO2_TD_V100
        if self.composit_type == 'TD':
            fname_date_tag = self.process_date
            first_day_of_period = self.process_date
            #eg : /data/TERRASCOPE/s5p/output_nc/L3_CO_TD_V1/2020/12/S5P_OFFL_L3_CO_TD_20201218_V100/
            self.outfolder_nc = os.path.join(output_dir_nc, self.process_date[0:4], self.process_date[4:6], 'S5P_%s_L3_%s_%s_%s_V%s' % (self.processing_mode, self.product_type, self.composit_type, first_day_of_period, self.version))
            self.outfolder_tif = os.path.join(output_dir_tif, self.process_date[0:4], self.process_date[4:6], 'S5P_%s_L3_%s_%s_%s_V%s' % (self.processing_mode, self.product_type, self.composit_type, first_day_of_period, self.version))
        elif self.composit_type == 'TM':
            fname_date_tag = self.process_date[0:6]             #eg: 202006
            first_day_of_period = '%s01' % fname_date_tag       #eg: 20200601 (first day of the month)
            #eg : /data/TERRASCOPE/s5p/output_nc/L3_CO_TM_V1/2022/S5P_OFFL_L3_CO_TM_20221001_V200/S5P_OFFL_L3TMCO___20221001T000000_20221031T235959_0001e_020400.nc
            self.outfolder_nc = os.path.join(output_dir_nc, self.process_date[0:4],
                                             'S5P_%s_L3_%s_%s_%s_V%s' % (
                                             self.processing_mode, self.product_type, self.composit_type,
                                             first_day_of_period, self.version))
            self.outfolder_tif = os.path.join(output_dir_tif, self.process_date[0:4],
                                              'S5P_%s_L3_%s_%s_%s_V%s' % (
                                              self.processing_mode, self.product_type, self.composit_type,
                                              first_day_of_period, self.version))
        elif self.composit_type == 'TY':
            fname_date_tag = self.process_date[0:4]  # eg: 2020
            first_day_of_period = '%s0101' % fname_date_tag  # eg: 20200101 (first day of the year)
            #/data/TERRASCOPE/s5p/output_nc/L3_CO_TY_V1/S5P_OFFL_L3_CO_TY_20200101_V100/S5P_OFFL_L3TYCO___20200101T000000_20201231T235959_0016e_01_010400.nc
            self.outfolder_nc = os.path.join(output_dir_nc,
                                             'S5P_%s_L3_%s_%s_%s_V%s' % (
                                             self.processing_mode, self.product_type, self.composit_type,
                                             first_day_of_period, self.version))
            self.outfolder_tif = os.path.join(output_dir_tif,
                                              'S5P_%s_L3_%s_%s_%s_V%s' % (
                                              self.processing_mode, self.product_type, self.composit_type,
                                              first_day_of_period, self.version))



        # copy the dict, (not make another pointer to the same one)
        self.tif_settings = copy.deepcopy(TIF_DICT[self.product_type])           #loading default settings for this product_type(CO,NO2,...)
        list(map(lambda x: x.update({'processingmode': self.processing_mode }), self.tif_settings))  #add processing_mode(NRTI/OFFL) to the settings

        out_tif_file = 'S5P_%s_%s_%s_<band>_V%s.tif' % (self.product_type, self.composit_type, fname_date_tag, self.version)        #eg S5P_CO_TM_202001_CO_V100.tif
        list(map(lambda x: x.update({'tif_file_name': out_tif_file}), self.tif_settings))       #update tif_settings with the tif_filename
        self.tif_file_lst = list(map(lambda x: out_tif_file.replace('<band>', x.get("tif_band_name")), self.tif_settings))      #filenames of the tifs; remove <band> from filenames and add the names to a list (will contain the variable + weigt band)

        self.outncfile = 'S5P_%s_L3%s%s___%s_%s_%s_%s.nc' %(
            self.processing_mode,
            self.composit_type,
            self.product_type,
            self.begin_position,
            self.end_position,
            ''.join('{:05x}'.format(self.nr_of_inputs)),        #number of used inputs in a 5 digit hex number
            self.out_detailed_version)                              # eg. S5P_OFFL_L3TDNO2___20201218T011358_20201219T001150_0000e_01_010400.nc
        self.outtiffile = self.outncfile.replace('.nc', '.tif') #use the same name for nc and tif
        self.logger.info("-I- outputfile : %s" % self.outtiffile)

    def _set_process_parameters(self):
        #Defaults
        oper = []
        exclude_default = 'exclude(latitude_bounds_weight,longitude_bounds_weight,latitude,longitude)'
        option = ''
        extra_harp_args = '--format netcdf --hdf5-compression 0 '
        keep = self.keep    #so we could copy/paste from terrascope3
        oper_bin = self.oper_bin    #so we could copy/paste from terrascope3

        self.logger.info("-I- producttype : %s" % self.product_type)

        ######## CO processing ###########
        if self.product_type == 'CO':
            self.nc_band_names = ['CO_column_volume_mixing_ratio_dry_air', 'CO_column_volume_mixing_ratio_dry_air_count']
            if self.processing_mode == 'OFFL':
                option = 'co=corrected'

            tvar = 'CO_column_number_density'
            oper.extend(['derive(CO_column_volume_mixing_ratio_dry_air {time} [ppbv])',
                         keep % ('CO_column_volume_mixing_ratio_dry_air'), self.oper_bin])
            # add filter, depends on processor
            if self.out_detailed_version >= '010202':
                oper.insert(0, 'CO_column_number_density_validity >=70')  # mid level clouds
            else:
                oper = ['solar_zenith_angle <80', 'sensor_zenith_angle <65', 'cloud_height <5000',
                        'validity !& 4458240',
                        'scan_subindex >1'] + oper  # 4458240 for SRON ... 1073739775 more strict, better at wollongong

        elif self.product_type == 'NO2':
            self.nc_band_names=['tropospheric_NO2_column_number_density']
            tvar = 'tropospheric_NO2_column_number_density'
            oper.extend([keep % (','.join((tvar, 'scan_subindex', 'cloud_fraction'))), oper_bin])
            oper = ['tropospheric_NO2_column_number_density_validity>75'] + oper

        elif self.product_type == 'HCHO':
            self.nc_band_names=['tropospheric_HCHO_column_number_density']
            tvar = 'tropospheric_HCHO_column_number_density'
            if self.composit_type == 'TD' or 'TM':
                option = 'amf=clear_sky'    #TODO only when using L2-data???
                oper.insert(0, 'scan_subindex >20;scan_subindex <430')  #Remove first 20 and last 20 values in every scanline – they are more noisy.
            oper.insert(0, '%s_validity >50' % tvar)
            oper.extend([keep % (tvar), oper_bin])

        elif self.product_type == 'CH4':
            self.nc_band_names = ['CH4_column_volume_mixing_ratio_dry_air']
            tvar = 'CH4_column_volume_mixing_ratio_dry_air'
            option = 'ch4=bias_corrected'   #TODO only at TD
            oper.extend([keep % (','.join((tvar,))), oper_bin])  # 'surface_albedo','cloud_fraction' --> for Mahesh
            oper.insert(0, 'CH4_column_volume_mixing_ratio_dry_air_validity >50')
        else:
            self.nc_band_names=['']
            self.logger.error("-E- non existing producttype")
            self.logger.error("-E- type : %s" % self.product_type)
            raise ValueError('not existing producttype : %s ' % self.product_type)

        if self.input_processing_level == 'L3':
            #spatial binning is removed here, make sure to use an input having the resolution as the requested output
            oper=[]
            oper.insert(0, exclude_default)

        if not isinstance(oper, str): oper = ';'.join(oper) #clean the list for empty strings, playing safe...
        self.harp_parameters = " %s %s -ar 'bin();exclude(latitude_bounds_weight,longitude_bounds_weight)' %s %s %s" % (
            ("-o '%s'" % option if option else ""),
            ("-a '%s'" % oper if oper else ""),
            extra_harp_args,
            self.infile_str,
            os.path.join(self.outfolder_nc, self.outncfile)
            )

    def _create_output_folder(self, output_dir_nc, output_dir_tif):
        set_umask()

        #creating output folders in advance
        if not os.path.exists(output_dir_nc):
            raise Exception("Output root folder does not exist, please create this before restart: %s" % output_dir_nc)
        if not os.path.exists(output_dir_tif):
            raise Exception("Output root folder does not exist, please create this before restart: %s" % output_dir_tif)


        #create output dir if not exist
        for folder in [self.outfolder_nc, self.outfolder_tif]:
            if not os.path.exists(folder):
                try:
                    os.makedirs(folder, exist_ok=True)
                    self.logger.info("-I- created folder %s " % folder)
                except:
                    self.logger.error("-E- unable to create dir %s" % folder)
                    return -1

    def print_result(self, output_type='json'):
        if output_type.lower() == 'json':
            result={}
            # "title": "S5P_NRTI_L3_CO_TD_20221204_V200",
            # "id": "S5P_NRTI_L3_CO_TD_20221204_V200",
            # "properties": {
            #     "eo:collection": "S5P_NRTI_L3_CO_TD_200",
            #     "producttype": "CO_TD",
            #     "processorversion": "020400",
            #     "endposition": "2022-12-04T23:59:59Z",
            #     "beginposition": "2022-12-04T00:00:00Z",
            #     "processingmode": "NRTI",
            #     "processingbaseline": "V200",
            #     "nrofinputs": "177",
            #     "ingestiondate": "2022-12-05T08:32:01.866Z",
            #     "instrumentshortname": "TROPOMI",
            #     "mission": "Sentinel",
            #     "processinglevel": "L3",
            #     "platformshortname": "S5P",
            #     "platformname": "Sentinel-5 Precursor",
            #     "format": "netCDF",
            #     "processing_status": {
            #         "processing_end": "2022-12-05T08:32:01.866Z",
            #         "status": "DONE"
            #     };
            now = datetime.now()
            result['title']= self.title
            result['id'] = self.title

            properties = {}         #empty dicts for nested structure
            processing_status = {}  #empty dicts for nested structure
            processing_status['processing_end'] = now.strftime('%Y-%m-%dT%H:%M:%S.000Z')  # 2022-12-05T08:32:01.866Z
            processing_status['status'] = 'TODO'
            properties['processing_status'] = processing_status

            if self.status == 0:
                #the following information is probably missing if status !=0
                properties['eo:collection'] = self.collection
                properties['producttype'] = '%s_%s' % (self.product_type, self.composit_type)      #"CO_TD"
                properties['composittype'] = self.composit_type
                properties['processorversion'] = self.out_detailed_version
                properties['endposition'] = datetime.strptime(self.end_position, '%Y%m%dT%H%M%S').strftime('%Y-%m-%dT%H:%M:%S.000Z')   #2022-12-04T23:59:59Z
                properties['beginposition'] = datetime.strptime(self.begin_position,'%Y%m%dT%H%M%S').strftime('%Y-%m-%dT%H:%M:%S.000Z')    #2022-12-04T23:59:59Z
                properties['processingmode'] = self.processing_mode
                properties['processingbaseline'] = 'V%s' % self.version
                properties['nrofinputs'] = self.nr_of_inputs
                properties['ingestiondate'] = now.strftime('%Y-%m-%dT%H:%M:%S.000Z')    #2022-12-05T08:32:01.866Z
                properties['instrumentshortname'] = 'TROPOMI'
                properties['mission'] = 'Sentinel'
                properties['processinglevel'] = self.processing_level
                properties['platformshortname'] = "S5P"
                properties['platformname'] = "Sentinel-5 Precursor"
                properties['format'] = "netCDF"
                processing_status['status'] = 'DONE'
                #created nested json
                #create input dict [{title:S5P_NRTI_L2__CO_____20221211T001600_20221211T002100_26738_03_020400_20221211T012528},{title:...}]
                infiles = self.infile_str.split(" ")
                inputs=[]
                for infile in infiles:
                    fname = os.path.basename(infile)
                    in_d = {}
                    in_d.update(title=fname.split(".")[0])     #remove fileextention
                    inputs.append(in_d)
                result['inputs'] = inputs
                outputs={}
                outputs['nc_file'] = self.outfolder_nc + '/' + self.outncfile
                outputs['tif_dir'] = self.outfolder_tif
                outputs['tif_files'] = self.tif_file_lst
                result['out'] = outputs
        result['properties'] = properties
        return result   #result is a dict here


def set_umask():
    """Set umask for right permissions"""
    try:
        os.umask(0o000)
    except:
        pass
