#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import logging
logformat='%(asctime)s %(name)s %(levelname)-8s %(message)s'
logtfmt='%y-%m-%d %H:%M:%S'
#set the basis configuration for logging to the console 
logging.basicConfig(level=logging.INFO,format=logformat,datefmt=logtfmt)
logger=logging.getLogger('BIRA')
testlogger=logging.getLogger('TEST')
testlogger.setLevel(logging.DEBUG)
operlogger=logging.getLogger('TERRA')
operlogger.setLevel(logging.ERROR)
import argparse,os,fnmatch,glob,re
import numpy as np
from scipy.io import netcdf
import h5py
import datetime as dt
from jinja2 import Environment, FileSystemLoader

TEMPLATE_PATH=os.path.dirname(__file__)
#JINJA
template_loader = FileSystemLoader(searchpath=TEMPLATE_PATH)
jinja_environment = Environment(loader=template_loader)

__version__='0.9'
_LENS5PID=len('S5P_RPRO_L2__CO____')
HDF5='HDF5' #how to describe the file mormat HDF5 binary ? or HDF
NETCDF3='NetCDF-3'

xmldatefmt='%Y-%m-%dT%H:%M:%SZ' #'2018-09-05T12:00:15Z'

INFO=dict(TD='daily',TM='monthly',TY='yearly',m5='Saterdays (montly)',m6='Sundays (monthly)',mw='weekdays (monthly)',ms='weekend days (monthly)')

def analyze_s5p_attrs(f,target,fdatefmt='%Y%m%dT%H%M%S',logger=logger):
  attrs={}
  if h5py.is_hdf5(f): ncfid=h5py.File(f,'r');ncatts=ncfid.attrs;attrs['fileformat']=HDF5;variables=ncfid.keys()
  else: ncfid=netcdf.netcdf_file(f,'r');ncatts=ncfid.__dict__;attrs['fileformat']=NETCDF3;variables=ncfid.variables.keys()
  logger.debug('all file attributes: %s'%ncatts.keys())
  targetvars=sorted(filter(lambda x: '_column' in x and target.lower() in x.lower(),variables),key=len)
  x=(ncfid.variables[targetvars[0]].data if attrs['fileformat'] in (NETCDF3,) else ncfid[targetvars[0]][...]).flatten()
  attrs['missingdatapercentage']='%.4f'%(100.*np.isnan(x).sum()/x.shape[0])
  header=str(ncatts['history'].decode('ascii'));headerlist=header.split()
  ncfid.close()
  headerlist.reverse()
  headerlist=headerlist[:headerlist.index('harpmerge')+3] #keep history of most recent call of harpmerge
  headerlist.reverse()
  logger.debug('Header history = %s...'%headerlist[:10])
  attrs['firstprocessingdatetime']=headerlist[0]
  attrs['harpversion']='V'+re.sub('[^0-9.]','',headerlist[1])
  #attrs['listofinputfiles']=list(filter(lambda x: 'L2__' in x,header))
  attrs['listofinputfiles']=[dict(basename=os.path.basename(f),size=os.path.getsize(f) if os.path.isfile(f) else np.nan) for f in list(filter(lambda _f: '_L3' in _f and target in _f, headerlist))[:-1]]
  for i in attrs['listofinputfiles']: i['starttime']=dt.datetime.strptime( i['basename'].split('_')[-6],fdatefmt).strftime(xmldatefmt)
  try: 
    parts=list(map(float,(header.rsplit('bin_spatial',1)[-1].split(')')[0].strip('(').split(','))))# ['(3601', '-90', '0.05', '7201', '-180', '0.05']
    attrs['latres']=parts[2];attrs['lonres']=parts[5]
    clat=(parts[0]-1);clon=(parts[3]-1)
    attrs['boundingbox']=[parts[4],parts[4]+clon*attrs['lonres'],parts[1],parts[1]+clat*attrs['latres']]
  except: #try the old definition
    grid=eval(header.rsplit('bin_spatial',1)[-1].split('\'')[0].strip());
    attrs['latres']=grid[0][1]-grid[0][0];attrs['lonres']=grid[1][1]-grid[1][0]
    attrs['boundingbox']=[min(grid[0]),max(grid[0]),min(grid[1]),max(grid[1])]
  logger.debug('Info: %s'%attrs)
  return attrs

def create_xml(f='/bira-iasb/data/SATELLITE/S5P/TROPOMI/L3/HARP/OFL/CO/2019/S5P_OFFL_L3TMCO_____20191001T012402_20191101T013945_001b8_01_010202_20191116T225124.nc',outputfolder=None,template='inspire_S5P_template.xml',fdatefmt='%Y%m%dT%H%M%S',logger=logger):
  """creates a meta data file using a template"""
  template_file=template
  template=jinja_environment.get_template(template)
  bname=os.path.basename(f)
  head=bname[:_LENS5PID]
  target=head[13:].replace('_','');logger.debug('Target is %s'%target)
  bintype=head[11:13]
  if '_L3' not in bname: raise ValueError('I only accept L3 files')
  fcomponents=bname[_LENS5PID:].rsplit('.',1)[0].split('_')
  attrs=analyze_s5p_attrs(f,target=target,fdatefmt=fdatefmt,logger=logger)
  gjsontitle=bname.rsplit('.',1)[0]
  pid='urn:eop:VITO:CGS_%s'%head.strip('_')
  gjsonid='urn:eop:VITO:CGS_S5_%s_L3'%target
  bname_min_proctime=head+'_'.join(fcomponents[:-1])
  item_dict=dict(eopidentifier='%s:%s'%(pid,bname_min_proctime), #urn:eop:VITO:CGS_S5P_L2_CO:S5P_OFFL_L2__CO_____20191211T004624_20191211T022754_11189_01_010302
                 gjsonidentifier='%s:%s'%(gjsonid,gjsontitle),#urn:eop:VITO:CGS_S5_CO_L3:S5P_OFFL_L3TMCO_____20190301T013534_20190401T001638_001b6_01_010202_20190716T160537
                 gjsontitle=gjsontitle,
                 eopparent=pid, #urn:eop:VITO:CGS_S5P_L2_CO
                 eopproductype=bname[9:13],
                 archivedatetime=dt.datetime.now().strftime(xmldatefmt),
                 processingdatetime=dt.datetime.strptime( fcomponents[-1],fdatefmt).strftime(xmldatefmt),
                 productreference=head,
                 title=bname_min_proctime, #S5P_OFFL_L2__CO_____20191211T004624_20191211T022754_11189_01_010302
                 productdescription=head[:11].replace('_',' ').strip()+' %s binned %s data, regridded at %.2fdegx%.2fdeg resolution'%(INFO.get(bintype,'NOTYPE'),target,attrs.pop('latres',0.),attrs.pop('lonres',0.)),
                 productversion=__version__,
                 size=os.path.getsize(f),
                 listofvitotiffs=['julliegeotifffilenames'],
                 starttime=dt.datetime.strptime( fcomponents[-6],fdatefmt).strftime(xmldatefmt), #2019-12-11T00:46:24Z
                 endtime=dt.datetime.strptime( fcomponents[-5],fdatefmt).strftime(xmldatefmt),
                 pythonversion='V'+__version__,
                 resolution='?',
                 opermode='?')
  #some derived values
  item_dict['startdate']=item_dict['starttime'].split('T')[0]#2019-12-11
  item_dict['gjsonfilepath']=dt.datetime.strptime( fcomponents[-6],fdatefmt).strftime('file:///data/MTDA/CGS_S5P/CGS_S5P_L3/%Y/%m/%d/S5P_%Y%m%d_CGS_V100/')+bname
  item_dict['previews']=[dict(href=item_dict['gjsonfilepath'].rsplit('.')[0]+'_QL.tiff',size=100,cat='QUICKLOOK')]
  #complete with header info from file
  item_dict.update(attrs)
  page = template.render(item_dict)
  if outputfolder:
    outputfolder=os.path.expanduser(outputfolder)
    if not os.path.isdir(outputfolder): os.makedirs(outputfolder)
    targetfile=os.path.join(outputfolder,'%s.%s'%(gjsontitle,template_file.rsplit('.')[-1]))
    with open(targetfile,'w') as fid: fid.write(page)
    logger.info('Written %s'%targetfile)
  #print(page)